local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "tm_player0"
local trafficNames = {"tm_traffic0", "tm_traffic1", "tm_traffic2", "tm_traffic3"}
local failMessages = {crash = "Yikes. Call for a tow truck!"}
local winMessages = {normal = "Welcome home!", veryFast = "You're a pretty fast driver! Now you have extra free time.", ultraFast = "Wha? You're super fast, that's an insane time! Kudos to you."}
local paintColors = {
{{0.9, 0.05, 0.04, 1.35}, {1, 0.3, 0, 1.3}, {0.95, 0.7, 0.1, 1.35}, {0.0, 0.2, 0.7, 1.3}, {0.15, 0.35, 0.6, 1.3}, {0.66, 0.655, 0.65, 1.3}}, 
{{0.65, 0.65, 0.65, 1.2}, {0.2, 0.3, 0.6, 1.2}, {0.68, 0.1, 0.11, 1.3}, {0.24, 0.48, 0.48, 1.2}, {0, 0.04, 0.13, 1.2}, {0.3, 0, 0, 1.2}}, 
{{0.85, 0.85, 0.85, 1.85}, {0.9, 0.05, 0.04, 1.3}, {1, 0.3, 0, 1.2}, {0.22, 0.37, 0.33, 1.3}, {0.24, 0.48, 0.48, 1.2}, {0.052, 0.52, 0.90, 1.3}}, 
{{0.2, 0.3, 0.6, 1.2}, {0.12, 0.28, 0.18, 1.2}, {0.3, 0.2, 0.12, 1.2}, {0, 0, 0, 1.3}, {0.18, 0.18, 0.18, 1.2}, {0.5, 0.5, 0.5, 1.3}}
} --Traffic colors; follow the curly brackets
local quatAngle = {quatFromDir(vec3(0, 1, 0), vec3(0, 0, 1)), quatFromDir(vec3(1, 0, 0), vec3(0, 0, 1)), quatFromDir(vec3(0, -1, 0), vec3(0, 0, 1)), quatFromDir(vec3(-1, 0, 0), vec3(0, 0, 1))} --Angles; 0, 90, 180, 270
local switchModes = {"switchFlee", "switchChase"}

local player = {}
local traffic = {}
local markerFinish = {}
local markerSuper = {}
local alpha = {}
local Traffic = {}

local queue = 0
local speedLimit = 0
local mainTimer = 0
local funMode = 0

local timeLevel = 100 --Time needed to reward player
local playerSmash = 1000000 --Amount of damage on the player car that will fail the scenario
local trafficSmash = 1000 --Amount of damage on a traffic car that will enable it to reinitialize
local minSpeed = 15 --Minimum speed limit for randomizer, in metres per second
local maxSpeed = 20 --Maximum speed limit for randomizer, in metres per second
local crawlSpeed = 3 --Almost as slow as a turtle
local trafficAggro = 0.9 --Aggression level of the traffic cars
local nearDist = 20 --Near distance, used for traffic passing and interactions
local farDist = 300 --Far distance, used for traffic despawning
local pointDist = 200 --Distance at a point from ahead of player vehicle; used for setting a spawn point
local readyDist = 100 --Normal distance, used for a variety of traffic and police logic
local farMultiplier = 3 --Spawn logic multipler for distant traffic cars; used with readyDist
local randomColors = false --If true, traffic car colors can get changed to anything!

function Traffic:new(obj, pos, speed, oldDamage, damage, state, role, far, timer)
	local self = setmetatable({}, Traffic)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.pos = pos or nil --Vehicle position
	self.speed = speed or nil --Vehicle speed behavior
	self.oldDamage = oldDamage or 0 --Vehicle damage points without player hit
	self.damage = damage or 0 --Vehicle damage points
	self.state = state or nil --Teleport states; none, normal, ready, set
	self.role = role or nil --Purpose of the traffic car; normal
	self.far = far or 1 --Multiplier for spawn mode; see below
	self.timer = timer or 0 --Internal logic timer
	return self
end

local function setSpawn(aVehicle) --Sets the new spawn point of a traffic car and rotates it accordingly
	local allow = true
	local spawnPoint = {}
	local spawnWay = math.random(1, 2)
	local spawnError = 0
	spawnPoint[1], spawnPoint[2], spawnPoint[3] = map.findClosestRoad(vec3(player.obj:getPosition()) - vec3(player.obj:getDirectionVector()) * math.min(farDist, pointDist + player.vel) * -1) --Spawn point is set to a certain distance in front of the player vehicle
	while not spawnPoint[1] do
		spawnError = spawnError + 1
		spawnPoint[1], spawnPoint[2], spawnPoint[3] = map.findClosestRoad(vec3(player.obj:getPosition()) - vec3(math.random() * 2 - 1, math.random() * 2 - 1, 0) * pointDist) --Spawn point retry in a random node in the area
		if spawnError >= 1000 then
			spawnPoint[1], spawnPoint[2], spawnPoint[3] = nil, nil, nil
		end
	end
	local spawnPos = map.getMap().nodes[spawnPoint[1]].pos
	local spawnRot = quatFromDir(map.getMap().nodes[spawnPoint[1]].pos - map.getMap().nodes[spawnPoint[2]].pos, vec3(0,0,1)) --Best way to get node rotation
	local objAngle = (spawnWay * 2) - 1
	local posAngle = spawnWay * 2
	local shiftUp = 0.5
	local shiftOut = 2
	local posOffset = nil
	local rotOffset = nil
	local colorIn = nil
	local colorValues = nil
	for i, v in ipairs(traffic) do
		if traffic[i].obj.name == aVehicle.obj.name then
			colorIn = paintColors[i][math.random(1, #paintColors[i])]
		end
		if (spawnPos:toPoint3F() - traffic[i].pos):len() <= nearDist * 2 or map.getMap().nodes[spawnPoint[1]].links[spawnPoint[2]].drivability < 0.3 then --Prevents spawn points inside other traffic cars or non roads
			allow = false
		end
	end
	if allow == true then
		posOffset = vec3(spawnPos:toPoint3F()) - vec3(0, 1, 0):rotated(spawnRot):rotated(quatAngle[posAngle]) * shiftOut
		rotOffset = quatFromDir(vec3(0, 1, 0):rotated(spawnRot):rotated(quatAngle[objAngle]), vec3(0, 0, 1))
		aVehicle.obj:setPositionRotation(posOffset.x, posOffset.y, posOffset.z + shiftUp, 0, 0, rotOffset.z, rotOffset.w + (math.random(1, 100) / 1000)) --The random addition fixes a spawn rotation init bug
		log("Z", "gull", aVehicle.obj.name.." teleported to "..spawnPos.x..", "..spawnPos.y..", "..spawnPos.z.."; road type: "..map.getMap().nodes[spawnPoint[1]].links[spawnPoint[2]].drivability)
		if aVehicle.damage >= trafficSmash then
			aVehicle.obj:queueLuaCommand("vehicleResetted(false)")
		end
		if randomColors == true then
			colorValues = {math.random(), math.random(), math.random(), math.random() + 1}
		else
			colorValues = {colorIn[1], colorIn[2], colorIn[3], colorIn[4]}
		end
		aVehicle.obj.color = ColorF(colorValues[1], colorValues[2], colorValues[3], colorValues[4]):asLinear4F()
		aVehicle.obj.colorPalette0 = aVehicle.obj.color
		aVehicle.state = "set"
		aVehicle.timer = 0
	end
end

local function finishScenario()
	if mainTimer <= timeLevel - 20 then
		scenario.finish({msg = winMessages.ultraFast})
	elseif mainTimer <= timeLevel then
		scenario.finish({msg = winMessages.veryFast})
	else
		scenario.finish({msg = winMessages.normal})
	end
end

local function onScenarioUIReady()
	player = {obj = nil, pos = nil, vel = nil, dist = nil, damage = nil, damageFactor = nil, timer = nil, winTimer = nil, failTimer = nil, success = nil}
	player.obj = scenetree.findObject(playerName)
	player.timer = 0
	player.winTimer = -1
	player.failTimer = 0
	player.success = true
	queue = 0
	mainTimer = 0
	funMode = 0
	for i, v in ipairs(trafficNames) do
		traffic[i] = Traffic:new(scenetree.findObject(trafficNames[i]), nil, 0, 0, 0, "normal", "normal", farMultiplier, 0) --obj, pos, speed, oldDamage, damage, state, role, far, timer
	end
	markerFinish = scenetree.findObject("markerFinish")
	markerSuper = scenetree.findObject("markerSuper")
	markerFinish.instanceColor = ColorF(0.8, 0.08, 0, 0.8):asLinear4F()
	markerSuper.instanceColor = ColorF(0.8, 0.08, 0, 1):asLinear4F()
end

local function onRaceStart()
	player.obj = scenetree.findObject(playerName)
	speedLimit = math.random(minSpeed, maxSpeed) --Creates a new speed limit for all traffic cars; makes restarts more interesting but prevents traffic cars from bumping into each other
	log("Z", "gull", "Speed limit: "..speedLimit)
	for i, v in ipairs(traffic) do
		traffic[i].obj:queueLuaCommand("ai.setAggression("..tostring(trafficAggro)..")")
		traffic[i].obj:queueLuaCommand("ai.driveInLane('on')")
		traffic[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
		traffic[i].obj:queueLuaCommand("ai.setSpeedMode('limit')")
		traffic[i].obj:queueLuaCommand("ai.resetLearning(true)")
		traffic[i].obj:queueLuaCommand("ai.setMode('random')")
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.damage = map.objects[map.objectNames[player.obj.name]].damage
	if mainTimer < 1 then --Workaround for an odd bug with zWeight where it won't get the proper value the first time
		player.obj:queueLuaCommand("obj:queueGameEngineLua('zWeight = ('..obj:calcBeamStats().total_weight..')')") --Creates a global variable with the player vehicle weight
	elseif mainTimer == 1 then
		if funMode == 0 then
			player.damageFactor = 1 + (zWeight / 4000) --Extra damage tolerance formula
		end
		log("Z", "gull", "Damage factor: "..player.damageFactor)
	end
	for i, v in ipairs(traffic) do
		traffic[i].pos = traffic[i].obj:getPosition()
		player.dist = (player.pos - traffic[i].pos):len()
		traffic[i].damage = map.objects[map.objectNames[traffic[i].obj.name]].damage
		if traffic[i].state == "set" then
			if traffic[i].timer >= 0.5 then --Any traffic timers, such as here, intentionally delay AI mode changes; experimental!
				if funMode == 0 then
					traffic[i].obj:queueLuaCommand("ai.setMode('random')")
				elseif funMode == 1 then
					traffic[i].obj:queueLuaCommand("ai.setMode('flee')")
				elseif funMode == 2 then
					traffic[i].obj:queueLuaCommand("ai.setMode('chase')")
				end
				traffic[i].speed = speedLimit
				traffic[i].state = "normal"
				traffic[i].far = 1
				traffic[i].timer = 0
			end
			traffic[i].timer = traffic[i].timer + raceTickTime
		end
		if player.dist <= nearDist then --If you pass a traffic car, prepare it for teleportation
			traffic[i].state = "ready"
			traffic[i].far = 1
			traffic[i].timer = 0
		end
		if player.dist <= readyDist / 2 then
			traffic[i].oldDamage = traffic[i].damage --//Add damage blame system
		end
		if mainTimer >= 1 then --Enables traffic spawning after a certain time
			if player.dist >= farDist and traffic[i].state == "normal" then --If you get really far away from a traffic car, prepare it for teleportation
				traffic[i].state = "ready"
				traffic[i].far = farMultiplier
				traffic[i].timer = 0
			end
			if queue > 0 then
				if player.dist >= (readyDist * traffic[i].far) and traffic[i].state == "ready" then
					if traffic[i].timer == 0 then
						traffic[i].obj:queueLuaCommand("ai.setMode('disabled')")
						traffic[i].far = 0
					end
					if player.dist >= readyDist and traffic[i].timer >= 0.5 and traffic[i].obj.name == traffic[queue].obj.name then
						setSpawn(traffic[i])
						queue = -4 --Artificial delay for next spawn test; prevents lag and stuff
					end
					traffic[i].timer = traffic[i].timer + raceTickTime
				end
			end
		end
		if traffic[i].damage >= trafficSmash * 2 then --Traffic cars try to stop driving if heavily damaged
			traffic[i].speed = trafficSmash * 10 / traffic[i].damage
		end
		if funMode ~= 0 then
			traffic[i].speed = 100
		end
	end
	queue = queue + 1 --This gets updated every race tick to prevent two cars from being teleported at the same time
	if queue > #trafficNames then
		queue = 1
	end
	if queue > 0 then --Processes only one traffic speed assignment per race tick
		traffic[queue].obj:queueLuaCommand("ai.setSpeed("..tostring(traffic[queue].speed)..")")
	end
	if player.damage >= trafficSmash * 10 then
		if player.vel <= crawlSpeed then
			player.failTimer = player.failTimer + raceTickTime
			if player.failTimer >= 1 + funMode + math.ceil(100000 / (player.damage / player.damageFactor)) / 2 then --Scenario fail depends on damage and stuck time
				player.success = false
				scenario.finish({failed = failMessages.crash})
			end
		else
			player.failTimer = 0
		end
	end
	if player.winTimer >= 0 then
		player.winTimer = player.winTimer + raceTickTime
		if player.winTimer >= 1 then
			finishScenario()
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name then 
			if string.byte(tostring(data.triggerName)) == 102 then --Prevents stupid trigger confusion with bus stops
				player.winTimer = 0
			end
			if data.triggerName == switchModes[1] then
				helper.flashUiMessage("Mode: Traffic Flee", 2, false)
				for i, v in ipairs(traffic) do
					traffic[i].obj:queueLuaCommand("ai.setAggression(1)")
					traffic[i].obj:queueLuaCommand("ai.driveInLane('off')")
					traffic[i].obj:queueLuaCommand("ai.setMode('flee')")
				end
				funMode = 1
			elseif data.triggerName == switchModes[2] then
				helper.flashUiMessage("Mode: Traffic Chase", 2, false)
				for i, v in ipairs(traffic) do
					traffic[i].obj:queueLuaCommand("ai.setAggression(1)")
					traffic[i].obj:queueLuaCommand("ai.driveInLane('off')")
					traffic[i].obj:queueLuaCommand("ai.setMode('chase')")
				end
				funMode = 2
			end
		end
	end
end

local function onPreRender(dt, dtSim)
	local markerFinishPos = nil
	local markerSuperPos = nil
	if mainTimer >= 1 then
		markerFinishPos = markerFinish:getPosition()
		markerSuperPos = markerSuper:getPosition()
		alpha[1] = math.max(0, math.min(0.8, (player.pos - markerFinish:getPosition()):len() / 80)) --Marker fades out as the player gets closer
		alpha[2] = math.max(0, math.min(1, ((player.pos - markerSuper:getPosition()):len() - 120) / 400))
		markerFinish:setPosition(vec3(markerFinishPos.x, markerFinishPos.y, markerFinishPos.z - 0.000001):toPoint3F()) --Refreshes marker object
		markerSuper:setPosition(vec3(markerSuperPos.x, markerSuperPos.y, markerSuperPos.z - 0.000001):toPoint3F())
		markerFinish.instanceColor = ColorF(0.8, 0.08, 0, alpha[1]):asLinear4F()
		markerSuper.instanceColor = ColorF(0.8, 0.08, 0, alpha[2]):asLinear4F()
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onBeamNGTrigger = onBeamNGTrigger
M.onPreRender = onPreRender

return M